import torch

def get_mask_from_lengths(lengths):
    max_length = torch.max(lengths).item()
    ids = torch.arange(0, max_length, device=lengths.device)
    mask = (ids < lengths.unsqueeze(1)).bool()

    return mask

def get_drop_frame_mask_from_lengths(lengths, drop_frame_rate):
    batch_size = lengths.size(0)
    max_len = torch.max(lengths).item()


#     mask = get_mask_from_lengths(lengths)
    ##

    max_length = torch.max(lengths).item()
    ids = torch.arange(0, max_length, device=lengths.device)
    mask = (ids < lengths.unsqueeze(1) -5 ).bool()

    ###

    drop_mask = torch.empty((batch_size, max_len), device=lengths.device).uniform_(0., 1.) < drop_frame_rate

    drop_mask = drop_mask.float() * mask

    return drop_mask

def drop_frame(mels, global_mean, mel_lengths, drop_frame_rate):
    drop_mask = get_drop_frame_mask_from_lengths(mel_lengths, drop_frame_rate)
    drop_mask = drop_mask.unsqueeze(1)
    global_mean = global_mean[None,:,None]

    dropped_mels = mels*(1 - drop_mask) + global_mean*drop_mask

    return dropped_mels



import numpy as np
import torch

import matplotlib.pyplot as plt
from matplotlib import figure

def plot_loss(losses, figname, log_dir):
    fig = figure.Figure(figsize=(20,12), tight_layout=True)
    axs = fig.subplots(4,1)
    for axis_idx, loss_name in enumerate(['total loss', 'mel loss', 'gate loss', 'alignment loss']):
        axs[axis_idx].plot([item[0] for item in losses], [item[axis_idx+1] for item in losses])
        axs[axis_idx].set_title(loss_name)
    
    fig.savefig('{}/{}.jpg'.format(log_dir, figname))
    plt.close()

def plot_output(model_outputs, sample_mel, sample_idx, i, log_dir):
    fig = figure.Figure(figsize=(20,12), tight_layout=True)
    axs = fig.subplots(4,1, gridspec_kw={'height_ratios': [3, 1, 2, 2]})

    # attention
    axs[0].imshow(model_outputs[-1][0].cpu().numpy().T, aspect='auto', origin='bottom', 
              interpolation='none')
    axs[0].set_title('attention')
    # gate
    axs[1].plot(torch.sigmoid(model_outputs[-2][0]).cpu().numpy())
    axs[1].set_title('gate')
    # mels
    mel_postnet_prediction = model_outputs[1][0].cpu().numpy()
    axs[2].imshow(mel_postnet_prediction, aspect='auto', origin='bottom', 
               interpolation='none')
    axs[2].set_title('mels')
    
    axs[3].imshow(sample_mel.cpu().numpy(), aspect='auto', origin='bottom', 
               interpolation='none')
    axs[3].set_title('target')
    
    fig.savefig('{}/{:03d}_{:06d}.jpg'.format(log_dir, sample_idx, i))
    plt.close()
    
    np.save('{}/mel_{:03d}_{:06d}.npy'.format(log_dir, sample_idx, i), mel_postnet_prediction)



def get_linear_lambda_with_warmup(num_warmup_steps, num_training_steps):
    
    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return 1.0
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    return lr_lambda



import os
import torch
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import LambdaLR
import numpy as np

from vietvoice.model.tacotron2 import Tacotron2, Tacotron2Loss
from vietvoice.loader import TextMelCollate, TextMelLoader, save_mel 
from vietvoice.utils.lr_decay import get_linear_lambda_with_warmup
from vietvoice.utils.visualize import plot_loss, plot_output
from vietvoice.utils.logger import Logger
from vietvoice.utils.download_model import download_weights, map_state_dict
from vietvoice.symbol import Symbol

class Trainer():
    def __init__(self, config):
        
        train_config = config['train_config']
        symbols_config = config['symbols_config']
        data_config = config['data_config']
        encoder_config = config['encoder_config']
        decoder_config = config['decoder_config']
        postnet_config = config['postnet_config']

        self.device = train_config['device']
        self.batch_size = train_config['batch_size']
        self.precomputed_mels = train_config['precomputed_mels']
        self.num_iters = train_config['num_iters']
        self.eval_steps = train_config['eval_steps']
        self.warmup_steps = train_config['warmup_steps']
        self.log_dir = train_config['log_dir']
        
        self.symbols = Symbol(**symbols_config)

        if self.precomputed_mels != None:
            print('Mels will load from precomputed mels insteads of compute directly from audio')
            if os.path.isdir(self.precomputed_mels):
                print('Warning !!! {} existed. Please remove if you want to recaculate again.'.format(self.precomputed_mels))
            else:
                save_mel(train_config['train_files'], train_config['precomputed_mels'], data_config)
                save_mel(train_config['test_files'], train_config['precomputed_mels'], data_config)

        collate_fn = TextMelCollate()

        trainset = TextMelLoader(train_config['train_files'], text_to_index_fn=self.symbols,
                                 precomputed_mels = train_config['precomputed_mels'],
                                **data_config)

        self.train_loader = DataLoader(trainset, num_workers=3, shuffle=True,
                                  sampler=None,
                                  batch_size=train_config['batch_size'], pin_memory=False,
                                  drop_last=True, collate_fn=collate_fn)

        self.train_iter = iter(self.train_loader)

        evalset = TextMelLoader(train_config['test_files'], text_to_index_fn=self.symbols,
                                precomputed_mels = train_config['precomputed_mels'],
                                **data_config)

        self.eval_loader = DataLoader(evalset, num_workers=1, shuffle=False,
                                  sampler=None,
                                  batch_size=train_config['batch_size'], pin_memory=False,
                                  drop_last=True, collate_fn=collate_fn)
        
        self.model = Tacotron2(n_mel_channels=data_config['n_mel_channels'], 
                  n_frames_per_step=data_config['n_frames_per_step'],
                  drop_frame_rate=train_config['drop_frame_rate'], 
                  global_mean=self.global_mean(data_config), 
                  n_symbols = len(self.symbols),
                  symbols_embedding_dim= symbols_config['symbols_embedding_dim'],
                  encoder_config=encoder_config, 
                  decoder_config=decoder_config, 
                  postnet_config=postnet_config)
        
        sample_batch = next(iter(self.eval_loader))
        sample_x, sample_y = self.model.parse_input(sample_batch, self.device)
        self.sample_text_inputs, self.sample_text_lengths, self.sample_mels, self.sample_output_lengths = sample_x

        if self.device.startswith('cuda'):
            self.model = self.model.to(self.device)
        
        self.criterion = Tacotron2Loss()
        self.optimizer = torch.optim.AdamW(self.model.parameters(), lr=train_config['learning_rate'],
                                 weight_decay=1e-6)

        self.lr_lambda = get_linear_lambda_with_warmup(self.warmup_steps, self.num_iters)
        self.scheduler = LambdaLR(self.optimizer, self.lr_lambda)

        self.best_eval_loss = 99999
        self.current_iter = 0
        self.train_losses = []
        self.eval_losses = []
        
        if self.log_dir != None:
            self.log = Logger(os.path.join(self.log_dir, 'log.txt'))
        
        pretrained = train_config['pretrained_url']
        if pretrained != None:
            print('Load pretrained model')
            pretrained_weights_path = download_weights(pretrained, md5=train_config['pretrained_md5'])
            state_dict = torch.load(pretrained_weights_path)
            state_dict = map_state_dict(self.model, state_dict)
            self.model.load_state_dict(state_dict, strict=False)

    def train(self):
        for i in range(self.current_iter, self.num_iters):
            self.current_iter = i
            try:
                batch = next(self.train_iter)
            except StopIteration:
                self.train_iter = iter(self.train_loader)
                batch = next(self.train_iter)
            
            x, y = self.model.parse_input(batch, self.device)
            
            loss_items = self.train_step(x, y, i)
            str_log = 'Iter: {:06d} - total loss: {:.6f} - mel loss: {:.6f} - gate loss: {:.6f} - alignemnt loss: {:.6f} - lr: {:.2e}'.format(i, *loss_items, self.optimizer.param_groups[0]['lr'])
            self.log(str_log)

            if i >= 1000:
                self.train_losses.append((i, *loss_items))
            
            if i % self.eval_steps == 0:
                eval_loss_items = self.evaluate()
                self.sample()

                self.eval_losses.append((i, *eval_loss_items))

                plot_loss(self.train_losses, 'train_loss', self.log_dir)
                plot_loss(self.eval_losses, 'eval_loss', self.log_dir)
                
                str_log = 'evaluation: iter: {:06d} - total loss: {:.6f} - mel loss: {:.6f} - gate loss: {:.6f} - alignemnt loss: {:.6f}'.format(i, *eval_loss_items)
                self.log(str_log)

                if eval_loss_items[0] < self.best_eval_loss:
                    self.best_eval_loss = eval_loss_items[0]
                    self.save_checkpoint()

    def train_step(self, x, y, i):
        self.model.train()
        self.model.zero_grad()
        guide_decay = self.lr_lambda(i)

        y_pred = self.model(x)
        
        
        loss, loss_items = self.criterion(y_pred, y, x[1], x[-1], attention_guide_decay=0)
        
        loss.backward()
        
        grad_norm = torch.nn.utils.clip_grad_norm_(
                        self.model.parameters(), 1.0)    
            
        self.optimizer.step()
        self.scheduler.step()

        return loss_items

    def evaluate(self):
        validation_losses = []
        self.model.eval()

        for batch in self.eval_loader:
            x, y = self.model.parse_input(batch, self.device)

            with torch.no_grad():
                y_pred = self.model(x)

            loss, loss_items = self.criterion(y_pred, y, x[1], x[-1], attention_guide_decay=0) 
            validation_losses.append(loss_items)

        avg_validation_loss = np.mean(validation_losses, axis=0).tolist()

        return avg_validation_loss
    
    def global_mean(self, data_config):
        global_mean = torch.zeros(data_config['n_mel_channels']*data_config['n_frames_per_step'])
        total_frame = 0

        for batch in self.train_loader:
            _, _, mel_padded, _, output_lengths, _ = batch
            global_mean += mel_padded.sum(dim=(0, 2))
            total_frame += output_lengths.sum()
            
        global_mean = global_mean/total_frame

        if self.device.startswith('cuda'):
            global_mean = global_mean.to(self.device)

        return global_mean

    def sample(self):
        self.model.eval()
        for k in range(len(self.sample_text_inputs)):
            sample_text_ids = self.sample_text_inputs[k][:self.sample_text_lengths[k]][None,:]
            sample_mel = self.sample_mels[k][:, :self.sample_output_lengths[k]]
            
            with torch.no_grad():
                sample_outs = self.model.inference(sample_text_ids)
            
            plot_output(sample_outs, sample_mel, k, self.current_iter, self.log_dir) 

    def save_checkpoint(self):
        state = {
            'iter': self.current_iter,
            'best_eval_loss': self.best_eval_loss,
            'state_dict': self.model.state_dict(),
            'optimizer': self.optimizer.state_dict(),
            'scheduler': self.scheduler.state_dict(),
            'train_losses': self.train_losses,
            'eval_losses': self.eval_losses
        }    
        
        torch.save(state, os.path.join(self.log_dir, 'checkpoint.pt'))
        torch.save(self.model.state_dict(), os.path.join(self.log_dir, 'model.pt'))

    def load_checkpoint(self):
        state = torch.load(os.path.join(self.log_dir, 'checkpoint.pt'))
        self.current_iter = state['iter']
        self.best_eval_loss = state['best_eval_loss'] 
        self.model.load_state_dict(state['state_dict'])
        self.optimizer.load_state_dict(state['optimizer'])
        self.scheduler.load_state_dict(state['scheduler'])
        self.train_losses = state['train_losses']
        self.eval_losses = state['eval_losses']

        print('continue train from the checkpoint {}. current_iter: {}, best_eval_loss: {}'.format(os.path.join(self.log_dir, 'checkpoint'), self.current_iter, self.best_eval_loss))



import torch
from scipy.io import wavfile
import math
from numba import jit
import numpy as np
import os

from vietvoice.stft import TacotronSTFT

MAX_WAV_VALUE = 32768.0

def load_wav_to_torch(full_path):
    sampling_rate, data = wavfile.read(full_path)
    return torch.FloatTensor(data.astype(np.float32)), sampling_rate


def load_filepaths_and_text(filename, split="|"):
    with open(filename, encoding='utf-8') as f:
        filepaths_and_text = [line.strip().split(split) for line in f]
    return filepaths_and_text

@jit
def get_guided_alignment(text_length, mel_length):
    w = np.zeros((mel_length, text_length))
    
    for i in range(mel_length):
        for j in range(text_length):
            w[i, j] = 1 - math.exp(-((i/mel_length - j/text_length)**2)/(2*0.2**2))
    
    return w

def save_mel(fin, out, data_config):
    stft = TacotronSTFT(**data_config)
    
    os.makedirs(out, exist_ok=True)

    filenames = load_filepaths_and_text(fin)

    for item in filenames:
        audiopath = item[0]
        audio, sampling_rate = load_wav_to_torch(audiopath)
        audio_norm = audio/MAX_WAV_VALUE
        audio_norm = audio_norm.unsqueeze(0)
        mel = stft.mel_spectrogram(audio_norm)
        mel = mel.squeeze(0).numpy()

        np.save('{}/{}.npy'.format(out, os.path.basename(audiopath)), mel)

class TextMelCollate():
    
    def __init__(self):
        pass
    
    def __call__(self, batch):
        
        batch_size = len(batch)
        text_lengths = [len(text) for text, _, _ in batch]
        
        input_lengths, ids_sorted_descreasing = torch.sort(
            torch.LongTensor(text_lengths),
            dim=0,
            descending=True
        )
    
        max_input_length = input_lengths[0]
        
        text_padded = torch.LongTensor(batch_size, max_input_length).zero_()
        
        for i in range(batch_size):
            text = batch[ids_sorted_descreasing[i]][0]
            text_padded[i, :len(text)] = text
        
        num_mels = batch[0][1].size(0)
        
        max_target_len = max([x[1].size(1) for x in batch])            
            
        mel_padded = torch.FloatTensor(batch_size, num_mels, max_target_len).zero_()        
        gate_padded = torch.FloatTensor(batch_size, max_target_len).zero_()
        output_lengths = torch.LongTensor(batch_size)
        
        for i in range(batch_size):
            mel = batch[ids_sorted_descreasing[i]][1]
            mel_padded[i, :, :mel.size(1)] = mel
            gate_padded[i, mel.size(1) - 1: ] = 1
            output_lengths[i] = mel.size(1)
        
        guided_alignments = torch.zeros(batch_size, max_target_len, max_input_length)
        for i in range(batch_size):
            alignment = batch[ids_sorted_descreasing[i]][2]
            guided_alignments[i,:alignment.size(0),:alignment.size(1)] = alignment
        
        return text_padded, input_lengths, mel_padded, gate_padded, output_lengths, guided_alignments
    
class TextMelLoader(torch.utils.data.Dataset):
    
    def __init__(self, dataset, text_to_index_fn, n_frames_per_step, sampling_rate, 
                 filter_length, hop_length, win_length, n_mel_channels, 
                 mel_fmin, mel_fmax, precomputed_mels=None):
        
        self.dataset = load_filepaths_and_text(dataset)
        self.text_to_index_fn = text_to_index_fn
        self.n_frames_per_step = n_frames_per_step
        
        self.precomputed_mels = precomputed_mels
        
        self.stft = TacotronSTFT(filter_length=filter_length, hop_length=hop_length, 
                         win_length=win_length, n_mel_channels=n_mel_channels, 
                         sampling_rate=sampling_rate, 
                         mel_fmin=mel_fmin, mel_fmax=mel_fmax)
            
    def __len__(self):
        return len(self.dataset)
    
    def get_mel_text(self, audiopath_and_text):
        audiopath, text = audiopath_and_text[0], audiopath_and_text[1]
        
        mel = self.get_mel(audiopath)
        text = self.get_text(text)
        
        return (text, mel)
    
    def get_text(self, text):
        text = self.text_to_index_fn(text)        
    
        text = torch.IntTensor(text)
        
        return text
    
    def get_mel(self, audiopath):
        audio, sampling_rate = load_wav_to_torch(audiopath)
        audio_norm = audio/MAX_WAV_VALUE
        audio_norm = audio_norm.unsqueeze(0)
        mel = self.stft.mel_spectrogram(audio_norm)
        mel = mel.squeeze(0)
        
        return mel
            
        
    def __getitem__(self, index):
        audiopath_and_text = self.dataset[index]
        audiopath, text = audiopath_and_text[0], audiopath_and_text[1]
        
        text = self.get_text(text)
        
        if self.precomputed_mels is None:
            mel = self.get_mel(audiopath)
        else:
            mel_fname = os.path.basename(audiopath)
            mel = torch.from_numpy(np.load('{}/{}.npy'.format(self.precomputed_mels, mel_fname)))
        
        if self.n_frames_per_step > 1:
            mel_length = mel.size(1)
            if mel_length % self.n_frames_per_step != 0:
                mel_length = (mel_length//self.n_frames_per_step)*self.n_frames_per_step
            new_mel = mel[:, :mel_length]
            
            mel = new_mel.transpose(0, 1)
            mel = mel.reshape(mel_length//self.n_frames_per_step, -1)
            mel = mel.transpose(0, 1)
            
        guided_alignment = get_guided_alignment(text.size(0), mel.size(1))
        guided_alignment = torch.from_numpy(guided_alignment)
        
        return text, mel, guided_alignment



import torch
import torch.nn as nn
import torch.nn.functional as F

from vietvoice.model.layers import Encoder, Decoder, Postnet
from vietvoice.utils.mask import get_mask_from_lengths, drop_frame

class Tacotron2Loss(nn.Module):
    def __init__(self):
        super(Tacotron2Loss, self).__init__()        
        self.attention_weights = 10
        
    @staticmethod
    def masked_l2_loss(out, target, mel_lengths):
        num_not_padded = mel_lengths.sum() * out.size(1)
        loss = F.mse_loss(out, target, reduction="sum")
        loss = loss / num_not_padded
    
        return loss
    
    @staticmethod
    def masked_alignment_loss(alignment, guided_alignment, text_lengths, mel_lengths):
        num_not_padded = torch.sum(text_lengths*mel_lengths)
        loss = alignment*guided_alignment
        loss = torch.sum(loss)/num_not_padded
        
        return loss
    
    def forward(self, model_outputs, targets, text_lengths, mel_lengths, attention_guide_decay):
        mel_targets, gate_targets, guided_alignment = targets
        
        mel_outputs, mel_outputs_postnet, gate_outputs, alignment = model_outputs
        
        mel_loss = self.masked_l2_loss(mel_outputs, mel_targets, mel_lengths) + self.masked_l2_loss(mel_outputs_postnet, mel_targets, mel_lengths)
    
        gate_loss = nn.BCEWithLogitsLoss()(gate_outputs, gate_targets)
        
        alignment_loss = self.masked_alignment_loss(alignment, guided_alignment, text_lengths, mel_lengths)
        alignment_loss = attention_guide_decay*self.attention_weights*alignment_loss
        
        total_loss = mel_loss + gate_loss + alignment_loss
                
        return total_loss, (total_loss.item(), mel_loss.item(), gate_loss.item(), alignment_loss.item())

class Tacotron2(nn.Module):
    def __init__(self, n_mel_channels, n_frames_per_step, drop_frame_rate, global_mean, n_symbols, symbols_embedding_dim, 
                 encoder_config, decoder_config, postnet_config):
        super(Tacotron2, self).__init__()
        
        self.n_mel_channels = n_mel_channels
        self.n_frames_per_step = n_frames_per_step
        self.drop_frame_rate = drop_frame_rate
        self.global_mean = global_mean
        
        self.embedding = nn.Embedding(n_symbols, symbols_embedding_dim, padding_idx=0)
        self.encoder = Encoder(**encoder_config)
        self.decoder = Decoder(n_mel_channels=n_mel_channels*n_frames_per_step,
                               **decoder_config)
        self.postnet = Postnet(n_mel_channels=n_mel_channels*n_frames_per_step,
                               **postnet_config)
    
    def parse_input(self, batch, device):
        text_padded, input_lengths, mel_padded, gate_padded, output_lengths, guided_alignments  = batch
        
        text_padded = text_padded.to(device)
        input_lengths = input_lengths.to(device)
        mel_padded = mel_padded.to(device)
        gate_padded = gate_padded.to(device)
        output_lengths = output_lengths.to(device)
        guided_alignments = guided_alignments.to(device)
        
        X = text_padded, input_lengths, mel_padded, output_lengths
        y = mel_padded, gate_padded, guided_alignments
        
        return  X, y
    
    def transform_output(self, mel_outputs, mel_outputs_postnet, gate_outputs, alignments, text_lengths=None, output_lengths=None):
        if output_lengths is not None:
            mask_t_in = get_mask_from_lengths(text_lengths)
            mask_t_out = get_mask_from_lengths(output_lengths)
            mask_alignments = mask_t_out.unsqueeze(-1)*mask_t_in.unsqueeze(1)
            
            mask_t_in = ~mask_t_in
            mask_t_out = ~mask_t_out
            mask_alignments = ~mask_alignments
            
            mask_output = mask_t_out.expand(self.n_mel_channels*self.n_frames_per_step, -1, -1)
            mask_output = mask_output.permute(1, 0, 2) # BxHxT
            
            mel_outputs.masked_fill_(mask_output, 0.0)
            mel_outputs_postnet.masked_fill_(mask_output, 0.0)
            gate_outputs.masked_fill_(mask_output[:, 0, :], 1e3)
            alignments.masked_fill_(mask_alignments, 0.0)
            
        return mel_outputs, mel_outputs_postnet, gate_outputs, alignments
    
    def decouple_frames(self, mel_outputs):
        batch_size = mel_outputs.size(0)
        
        mel_outputs = mel_outputs.transpose(1, 2)
        mel_outputs = mel_outputs.reshape(batch_size, -1, self.n_mel_channels)
        mel_outputs = mel_outputs.transpose(1, 2)
        
        return mel_outputs
    
    def forward(self, inputs):
        text_inputs, text_lengths, mels, output_lengths = inputs
        
#         if self.global_mean != None and self.drop_frame_rate > 0 and self.training:
#             mels = drop_frame(mels, self.global_mean, output_lengths, self.drop_frame_rate)
         
        embedded_inputs = self.embedding(text_inputs)
        embedded_inputs = embedded_inputs.transpose(1, 2) # BxHxT
                
        encoder_outputs = self.encoder(embedded_inputs, text_lengths)
        
        mel_outputs, gate_outputs, alignments = self.decoder(encoder_outputs, mels, memory_lengths=text_lengths)
        
        mel_outputs_postnet = self.postnet(mel_outputs)
        mel_outputs_postnet = mel_outputs_postnet + mel_outputs
        
        outputs = self.transform_output(mel_outputs, mel_outputs_postnet, gate_outputs, alignments, text_lengths, output_lengths)
        
        return outputs
    
    def inference(self, inputs):
        """
        inputs: tensor of text idx
        """
        
        embedded_inputs = self.embedding(inputs).transpose(1, 2)
        encoder_outputs = self.encoder.inference(embedded_inputs)
        
        mel_outputs, gate_outputs, alignments = self.decoder.inference(encoder_outputs)
        
        mel_outputs_postnet = self.postnet(mel_outputs)
        mel_outputs_postnet = mel_outputs_postnet + mel_outputs
        
        outputs = self.transform_output(mel_outputs, mel_outputs_postnet, gate_outputs, alignments)
        return outputs



class Symbol:
    def __init__(self, symbols, kind = 'phoneme', delimiter='/', **kwargs):
        self.symbols = ['_PAD_'] + symbols
        self.symbol_to_id = {v:k for k, v in enumerate(self.symbols)}
        assert(kind in ['phoneme', 'char'])
        self.kind = kind
        self.delimiter=delimiter

    def __len__(self):
        return len(self.symbols)
    
    def __call__(self, text):
        if self.kind == 'phoneme':
            return self.phoneme_to_index(text)
        elif self.kind == 'char':
            return self.text_to_index(text)

    def phoneme_to_index(self, text):
        text = text.split(self.delimiter)
        text = [self.symbol_to_id[i] for i in text if len(i) > 0]
        return text

    def text_to_index(self, text):
        text = [self.symbol_to_id[i] for i in text]
        return text



import numpy as np
import torch

from vietvoice.model.tacotron2 import Tacotron2
from vietvoice.symbol import Symbol

class Predictor():
    def __init__(self, config, weights, device):

        symbols_config = config['symbols_config']
        data_config = config['data_config']
        encoder_config = config['encoder_config']
        decoder_config = config['decoder_config']
        postnet_config = config['postnet_config']

        self.device = device
        self.n_mel_channels = data_config['n_mel_channels']
        
        self.symbols = Symbol(**symbols_config)

        self.model = Tacotron2(n_mel_channels=self.n_mel_channels,
                  n_frames_per_step=data_config['n_frames_per_step'],
                  drop_frame_rate=0,
                  global_mean=None,
                  n_symbols = len(self.symbols),
                  symbols_embedding_dim= symbols_config['symbols_embedding_dim'],
                  encoder_config=encoder_config,
                  decoder_config=decoder_config,
                  postnet_config=postnet_config)
        self.model.eval()
        self.model.load_state_dict(torch.load(weights))
        
        if self.device.startswith('cuda'):
            self.model = self.model.to(self.device)

    def predict(self, text):
        indices = self.symbols(text)

        indices = torch.IntTensor(indices).reshape(1, -1)
        if self.device.startswith('cuda'):
            indices = indices.to(self.device)
        
        with torch.no_grad():
            outputs = self.model.inference(indices)
        
        mel = outputs[1].cpu().numpy()[0]
        # decouple for multi-frame per step
        mel = mel.T.reshape(-1, self.n_mel_channels).T

        return mel




import gdown
import requests
import yaml

def download_weights(id_or_url, cached=None, md5=None, quiet=False):
    if id_or_url.startswith('http'):
        url = id_or_url
    else:
        url = 'https://drive.google.com/uc?id={}'.format(id_or_url)

    return gdown.cached_download(url=url, path=cached, md5=md5, quiet=quiet)

def download_config(id):
    url = 'https://raw.githubusercontent.com/pbcquoc/config/master/vietvoice/{}.yml'.format(id)
    r = requests.get(url)
    config = yaml.safe_load(r.text)
    return config

def map_state_dict(model, state_dict):
    new_state_dict = {}
    required_state_dict = model.state_dict()
    
    for k, v in required_state_dict.items():
        if k in state_dict:
            if v.size() == state_dict[k].size():
                new_state_dict[k] = state_dict[k]
            else:
                print('pretrained weight dont match current weight shape {}'.format(k))
        else:
            print('not found weight for {}'.format(k))

    return new_state_dict



import torch
import torch.nn as nn
import torch.nn.functional as F

from vietvoice.utils.mask import get_mask_from_lengths

class ConvNorm(nn.Module):
    def __init__(self, in_channels, out_channels, 
                 kernel_size, stride=1, padding=None, dilation=1, bias=True, w_init_gain='linear'):
        super(ConvNorm, self).__init__()
        
        if padding is None:
            assert kernel_size % 2 ==1
            padding = dilation*(kernel_size - 1)//2
        
        self.conv = nn.Conv1d(in_channels, out_channels, kernel_size=kernel_size, 
                              stride=stride, padding=padding, dilation=dilation, bias=bias)    
        
        
        nn.init.xavier_uniform_(self.conv.weight, gain=nn.init.calculate_gain(w_init_gain))
        
        
    def forward(self, x):
        x = self.conv(x)
        
        return x
    
class Encoder(nn.Module):
    def __init__(self, n_convolutions, kernel_size, embedding_dim):
        super(Encoder, self).__init__()
        
        assert embedding_dim % 2 == 0
        
        self.convolutions = nn.ModuleList()
        for _ in range(n_convolutions):
            conv_norm = ConvNorm(embedding_dim, embedding_dim,
                                kernel_size = kernel_size, stride=1,
                                padding = (kernel_size -1) //2,
                                dilation=1, w_init_gain='relu'
                                )
            
            batch_norm = nn.BatchNorm1d(embedding_dim)
            
            conv_layer = nn.Sequential(
                conv_norm,
                batch_norm                
            )
            
            self.convolutions.append(conv_layer)
        
        self.lstm = nn.LSTM(embedding_dim, embedding_dim//2, 1, batch_first=True, bidirectional=True)
        
    
    def forward(self, x, input_lengths):
        """
        x: BxHxT
        """
        
        for conv in self.convolutions:
            x = F.dropout(F.relu(conv(x)), 0.5, self.training)
        
        x = x.transpose(1, 2)
        
        input_lengths = input_lengths.cpu().numpy()
        
        x = nn.utils.rnn.pack_padded_sequence(x, input_lengths, batch_first=True)
        
        outputs, _ = self.lstm(x)
        
        outputs, _ = nn.utils.rnn.pad_packed_sequence(outputs, batch_first=True)
        
        return outputs
        
    def inference(self, x):
        for conv in self.convolutions:
            x = F.dropout(F.relu(conv(x)), 0.5, self.training)
    
        x = x.transpose(1, 2)
        
        outputs, _ = self.lstm(x)

        return outputs

class LinearNorm(nn.Module):
    def __init__(self, in_dim, out_dim, bias=True, w_init_gain='linear'):
        super(LinearNorm, self).__init__()
        
        self.linear_layer = nn.Linear(in_dim, out_dim, bias=bias)
        nn.init.xavier_uniform_(self.linear_layer.weight, gain=nn.init.calculate_gain(w_init_gain))
        
    def forward(self, x):
        x = self.linear_layer(x)
        
        return x    
    
class Prenet(nn.Module):
    def __init__(self, in_dim, sizes):
        super(Prenet, self).__init__()
        in_sizes = [in_dim] + sizes[:-1]
        self.layers = nn.ModuleList()
        
        for in_size, out_size in zip(in_sizes, sizes):
            linear = LinearNorm(in_size, out_size, bias=False)
            
            self.layers.append(linear)
    
    def forward(self, x):
        for linear in self.layers:
            x = F.dropout(F.relu(linear(x)), 0.1, training=True)
        
        return x

class LocationLayer(nn.Module):
    def __init__(self, attention_n_filters, attention_kernel_size, attention_dim):
        super(LocationLayer, self).__init__()
        
        padding = (attention_kernel_size - 1)//2
        self.location_conv = ConvNorm(2, attention_n_filters, kernel_size=attention_kernel_size, padding=padding, bias=False, stride=1, dilation=1)
        
        self.location_dense = LinearNorm(attention_n_filters, attention_dim, bias=False, w_init_gain='tanh')
        
    def forward(self, attention_weights_cat):
        """
        attention_weights_cat: Bx2xT
        """
        processed_attention = self.location_conv(attention_weights_cat)
        
        processed_attention = processed_attention.transpose(1, 2)
        processed_attention = self.location_dense(processed_attention)
        
        return processed_attention
    
class Attention(nn.Module):
    def __init__(self, attention_rnn_dim, memory_dim, attention_dim,
                attention_location_n_filters, attention_location_kernel_size):
        super(Attention, self).__init__()
        
        self.query_layer = LinearNorm(attention_rnn_dim, attention_dim, bias=False, w_init_gain='tanh')
        self.memory_layer = LinearNorm(memory_dim, attention_dim, bias=False, w_init_gain='tanh')
        self.v = LinearNorm(attention_dim, 1, bias=False)
        
        self.location_layer = LocationLayer(attention_location_n_filters, attention_location_kernel_size, attention_dim)
        
        self.score_mask_value = -float('inf')
        
    def get_alignment_energies(self, query, processed_memory, attention_weights_cat):
        """
        query: BxAttention_rnn_dim
        process_memory: BxTxAttention_dim
        attention_weights_cat: cumulative and prev. att weights (B, 2, max_time)
        """
        
        query = query.unsqueeze(1)
        processed_query = self.query_layer(query)
        
        processed_attention_weights = self.location_layer(attention_weights_cat)
        
        energies = torch.tanh(processed_query + processed_attention_weights + processed_memory)
        energies = self.v(energies)
        energies = energies.squeeze(-1)
        
        return energies
    
    def forward(self, attention_hidden_state, memory, processed_memory, attention_weights_cat, mask):
        
        alignment = self.get_alignment_energies(attention_hidden_state, processed_memory, attention_weights_cat)
        
        if mask is not None:
            alignment.masked_fill_(mask, self.score_mask_value)
        
        
        attention_weights = F.softmax(alignment, dim=1)
        attention_context = torch.bmm(attention_weights.unsqueeze(1), memory)
        attention_context = attention_context.squeeze(1)
        
        return attention_context, attention_weights
    
class Decoder(nn.Module):
    def __init__(self, n_mel_channels, encoder_embedding_dim, 
                 attention_rnn_dim, decoder_rnn_dim, prenet_dim, 
                 max_decoder_steps, gate_threshold, 
                 p_attention_dropout, p_decoder_dropout, attention_config):
        super(Decoder, self).__init__()
        self.n_mel_channels = n_mel_channels
        self.encoder_embedding_dim = encoder_embedding_dim
        self.attention_rnn_dim = attention_rnn_dim
        self.decoder_rnn_dim = decoder_rnn_dim
        
        self.max_decoder_steps = max_decoder_steps
        self.gate_threshold = gate_threshold
        
        self.p_attention_dropout = p_attention_dropout
        self.p_decoder_dropout = p_decoder_dropout
        
        self.prenet = Prenet(n_mel_channels, [prenet_dim, prenet_dim])
        
        self.attention_rnn = nn.LSTMCell(prenet_dim + encoder_embedding_dim, attention_rnn_dim)
        
        self.attention_layer = Attention(attention_rnn_dim, encoder_embedding_dim, **attention_config)
        
        self.decoder_rnn = nn.LSTMCell(attention_rnn_dim + encoder_embedding_dim, decoder_rnn_dim)
        
        self.linear_projection = LinearNorm(decoder_rnn_dim + encoder_embedding_dim, n_mel_channels)
        
        self.gate_layer = LinearNorm(attention_rnn_dim + encoder_embedding_dim, 1, bias=True, w_init_gain='sigmoid')
    
    def get_go_frame(self, memory):
        batch_size = memory.size(0)
        
        decoder_input = memory.new(batch_size, self.n_mel_channels).zero_()
        return decoder_input
    
    def initialize_decoder_states(self, memory, mask):
        batch_size = memory.size(0)
        max_time = memory.size(1)
        
        # init state for first rnn
        self.attention_hidden = memory.new(batch_size, self.attention_rnn_dim).zero_()
        self.attention_cell = memory.new(batch_size, self.attention_rnn_dim).zero_()
        
        # init state for second rnn
        self.decoder_hidden = memory.new(batch_size, self.decoder_rnn_dim).zero_()
        self.decoder_cell = memory.new(batch_size, self.decoder_rnn_dim).zero_()
        
        # init attention
        self.attention_weights = memory.new(batch_size, max_time).zero_()
        self.attention_weights_cum = memory.new(batch_size, max_time).zero_()
        self.attention_context = memory.new(batch_size, self.encoder_embedding_dim).zero_()
        
        # save memory
        self.memory = memory
        self.processed_memory = self.attention_layer.memory_layer(memory)
        self.mask = mask
    
    
    def transform_decoder_outputs(self, mel_outputs, gate_outputs, alignments):

        # (T_out, B, n_mel_channels) -> # (B, n_mel_channels, T_out)
        mel_outputs = torch.stack(mel_outputs).permute(1, 2, 0)
        
        # (T_out, B) -> (B, T_out)
        gate_outputs = torch.stack(gate_outputs).transpose(0, 1)
         # (T_out, B, T_in) -> (B, T_out, T_in)
        alignments = torch.stack(alignments).transpose(0, 1)

        
        return mel_outputs, gate_outputs, alignments
    
    def decode(self, decoder_input):
        """
        decoder_input: previous mel output
        """
        
        # first rnn use attention_context of previous timestep
        cell_input = torch.cat((decoder_input, self.attention_context), dim=1)
        self.attention_hidden, self.attention_cell = self.attention_rnn(cell_input, 
                                                        (self.attention_hidden, self.attention_cell))
        
        self.attention_hidden = F.dropout(self.attention_hidden, self.p_attention_dropout, training=self.training)
        
        attention_weights_cat = torch.cat((
            self.attention_weights.unsqueeze(1), 
            self.attention_weights_cum.unsqueeze(1)),
            dim=1)
        
        # compute attention_context of current step
        self.attention_context, self.attention_weights = self.attention_layer(self.attention_hidden, self.memory, self.processed_memory, attention_weights_cat, self.mask)
        
        self.attention_weights_cum += self.attention_weights
        
        # second rnn
        decoder_input = torch.cat((self.attention_hidden, self.attention_context), dim=1)
        
        self.decoder_hidden, self.decoder_cell = self.decoder_rnn(decoder_input, 
                                                                  (self.decoder_hidden, self.decoder_cell))
        
        self.decoder_hidden = F.dropout(self.decoder_hidden, self.p_decoder_dropout, training=self.training)
        
        
        # linear projection 
        decoder_hidden_attention_context = torch.cat((self.decoder_hidden, self.attention_context), dim=1)
        
        decoder_output = self.linear_projection(decoder_hidden_attention_context)
        
        # gate prediction
        gate_prediction = self.gate_layer(decoder_hidden_attention_context)
        
        return decoder_output, gate_prediction, self.attention_weights
    
    def forward(self, memory, decoder_inputs, memory_lengths):
        """
        memory: encoder outputs
        decoder_inputs: decoder inputs for teaching force. i.e. mel-specs: BxHxT
        memory_lengths: encoder output lengths for attention masking.
        """
        # add init frame to decoder_inputs
        init_frame = self.get_go_frame(memory).unsqueeze(0) # 1xBxH
        decoder_inputs = decoder_inputs.permute(2, 0, 1) # TxBxH
        decoder_inputs = torch.cat((init_frame, decoder_inputs), dim=0)
        
        decoder_inputs = self.prenet(decoder_inputs)
        
        self.initialize_decoder_states(memory, mask= ~get_mask_from_lengths(memory_lengths))
        
        mel_outputs, gate_outputs, alignments = [], [], []        
        max_time = decoder_inputs.size(0) - 1
        
        for t in range(max_time):
            decoder_input = decoder_inputs[t]
            
            mel_output, gate_output, attention_weights = self.decode(decoder_input)            
                            
            mel_outputs.append(mel_output)
            gate_outputs.append(gate_output.squeeze(1))
            alignments.append(attention_weights)
            
        outputs = self.transform_decoder_outputs(mel_outputs, gate_outputs, alignments)
        
        return outputs
    
    def inference(self, memory):
        """
        memory: BxTxH
        """
        
        decoder_input = self.get_go_frame(memory)
        self.initialize_decoder_states(memory, mask=None)
        
        mel_outputs, gate_outputs, alignments = [], [], []
        
        while True:
            decoder_input = self.prenet(decoder_input)
            
            mel_output, gate_output, alignment = self.decode(decoder_input)
            mel_outputs.append(mel_output)
            gate_outputs.append(gate_output.squeeze(1))
            alignments.append(alignment)
            
            if torch.sigmoid(gate_output) > self.gate_threshold:
                break
            elif len(mel_outputs) >= self.max_decoder_steps:
                print("Warning! Reached max decoder steps")
                break
            
            decoder_input = mel_output
        
        outputs = self.transform_decoder_outputs(mel_outputs, gate_outputs, alignments)
        
        return outputs

class Postnet(nn.Module):
    def __init__(self, n_convolutions, n_mel_channels, embedding_dim, kernel_size): 
        super(Postnet, self).__init__()
        
        self.convolutions = nn.ModuleList()
        
        sizes = [n_mel_channels] + [embedding_dim]*(n_convolutions - 1) + [n_mel_channels]
        activations = ['tanh']*4 + ['linear']
        
        # output_size of previous conv is input_size of next layer
        for i in range(n_convolutions):
            padding = (kernel_size - 1)//2
            conv_norm = ConvNorm(sizes[i], sizes[i+1], kernel_size=kernel_size, padding=padding, stride=1, dilation=1, w_init_gain=activations[i])
            
            conv = nn.Sequential(
                conv_norm,
                nn.BatchNorm1d(sizes[i+1])
            )
            
            self.convolutions.append(conv)
    
    def forward(self, x):
        for i in range(len(self.convolutions) - 1):
            x = F.dropout(torch.tanh(self.convolutions[i](x)), 0.5, training=self.training)
        
        x = F.dropout(self.convolutions[-1](x), 0.5, training=self.training)
        
        return x            



import argparse
import numpy as np

from vietvoice.model.predictor import Predictor
from vietvoice.utils.config import Cfg

from vietvoice.vitext_to_phoneme import vitext_clean, vitext_to_phoneme

def predict(text, weights, out_file, device):
    config = Cfg.load_config_from_name('tacotron2_vi_phoneme')

    predictor = Predictor(config, weights=weights, device=device)
    
    text = vitext_clean(text)
    print('clean text: {}'.format(text))
    phoneme = vitext_to_phoneme(text)
    print('phoneme: {}'.format(phoneme))

    mel = predictor.predict(phoneme)
    
    np.save(args.out_file, mel)

if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--text', required=True, help='foo help')
    parser.add_argument('--weights', required=True, help='foo help')
    parser.add_argument('--device', default='cuda:0', help='foo help')
    parser.add_argument('--out_file', default='out.npy', help='foo help')

    args = parser.parse_args()

    predict(args.text, args.weights, args.out_file, args.device)



import string
import re

from underthesea import word_tokenize
from viphoneme import T2IPA_split

alphabet = 'abcdefghijklmnopqrstuvwxyzàáâãèéêìíòóôõùúýăđĩũơưạảấầẩẫậắằẳẵặẹẻẽếềểễệỉịọỏốồổỗộớờởỡợụủứừửữựỳỵỷỹ'
alphabet += string.punctuation
alphabet += ' '
alphabet = set(alphabet)

mapping = {'à':'à',
 'á':'á',
 'ã':'ã',
 'ả':'ả',
 'ạ':'ạ',
 'è':'è',
 'é':'é',
 'ẽ':'ẽ',
 'ẻ':'ẻ',
 'ì':'ì',
 'í':'í',
 'ĩ':'ĩ',
 'ỉ':'ỉ',
 'ị':'ị',
 'ò':'ò',
 'ó':'ó',
 'õ':'õ',
 'ỏ':'ỏ',
 'ọ':'ọ',
 'ù':'ù',
 'ú':'ú',
 'ũ':'ũ',
 'ủ':'ủ',
 'ụ':'ụ',
 'ỳ':'ỳ',
 'ý':'ý',
 'ỷ':'ỷ',
 'ầ':'ầ',
 'ấ':'ấ',
 'ẫ':'ẫ',
 'ẩ':'ẩ',
 'ậ':'ậ',
 'ề':'ề',
 'ế':'ế',
 'ễ':'ễ',
 'ể':'ể',
 'ệ':'ệ',
 'ồ':'ồ',
 'ố':'ố',
 'ỗ':'ỗ',
 'ổ':'ổ',
 'ộ':'ộ',
 'ằ':'ằ',
 'ắ':'ắ',
 'ẳ':'ẳ',
 'ặ':'ặ',
 'ờ':'ờ',
 'ớ':'ớ',
 'ỡ':'ỡ',
 'ở':'ở',
 'ợ':'ợ',
 'ừ':'ừ',
 'ứ':'ứ',
 'ữ':'ữ',
 'ử':'ử',
 'ự':'ự'}

_whitespace_re = re.compile(r'\s+')
def collapse_whitespace(text):
    return re.sub(_whitespace_re, ' ', text)

def vitext_clean(text):
    text  = text.strip()            
    text = text.lower()
    text = collapse_whitespace(text)
    text = text.replace('–', '-')
    
    text = re.sub('([.,!?()])', r' \1 ', text)
    text = re.sub('\s{2,}', ' ', text)

    for k, v in mapping.items():
        text = text.replace(k, v)
    
    for k in string.punctuation:
        if k not in [',', '.']:
            text = text.replace(k, ',')    
	
    return text

def vi2IPA(text):
    TK= word_tokenize(text)
    IPA=[]
    for tk in TK:
        ipa = T2IPA_split(tk, '/').replace(" ","_")        
        if ipa =="":
            IPA.append('/'+tk+'/')
        elif ipa[0]=="[" and ipa[-1]=="]":
            return []
        else:
            IPA.append(ipa)
        
    IPA = ' '.join(IPA)
    return IPA

def vitext_to_phoneme(text):
    phoneme = None
    if len(text) > 0 and set(text) <= alphabet:
        phoneme = vi2IPA(text)

    if len(text) == 0:
        print('text after cleaning is empty')

    if len(set(text) - alphabet) > 0:
        print('text contains invalid characters: {}'.format(set(text) - alphabet))

    return phoneme
